import java.util.*;
import java.awt.*;

/**
 * Esta clase representa a un chorro, visto como un sistema de particulas, que intenta
 * simular el comportamiente de un chorro de agua.
 * @author Ricardo Monascal
 **/
public class Chorro {

// Constantes de la clase.

	/**
	 * Gravedad que afecta a los chorros.
	 **/
	private static final double G = 1.0;

	/**
	 * Desplazamiento posible para las gotas.
	 **/
	private static final double EPS = 0.25;
	
	/**
	 * Cantidad mxima de partculas virtuales.
	 **/
	private static final int MAX_VIRTUALES = 2;

	/**
	 * Mxima cantidad de gotas posbles.
	 **/
	private static final int MAX_GOTAS = 60;

// Campos de la clase.

	/**
	 * Chorros de la fuente.
	 **/
	private Gota[] gotas;
	
	/**
	 * Color de este chorro (indice en el arreglo de colores del mundo de juego).
	 **/
	private int colorChorro;
	
	/**
	 * Radio base de las gotas del chorro.
	 **/
	private double radioChorro;
	
	/**
	 * Posicin inicial del chorro.
	 **/
	private Vector2D posChorro;
	
	/**
	 * Direccin principal del chorro (magnitud includa).
	 **/
	private Vector2D dirChorro;
	
	/**
	 * Indice de activacin / reseteo de gotas.
	 **/
	private int indiceGotas;

// Constructor de la clase.

	/**
	 * Constructor de la clase.
	 * @param   dir   Posicin de este chorro de agua.
	 * @param   dir   Direccin de este chorro de agua.
	 * @param   radio   Radio base de las gotas del chorro.
	 * @param   color   Color de las gotas del chorro.
	 **/
	protected Chorro (Vector2D pos, Vector2D dir, double radio, int color) {
		
		int i;
		
		gotas = new Gota [MAX_GOTAS];
		posChorro = pos;
		dirChorro = dir;
		radioChorro = radio;
		colorChorro = color;
		indiceGotas = 0;
		
		for (i=0 ; i<MAX_GOTAS ; i++)
			gotas[i] = new Gota();
		
	}

// Funciones y metodos de la clase.

	/**
	 * Asigna un nuevo color a este chorro.
	 * @param   color   El nuevo color del chorro.
	 **/
	public void asignarColor (int color) {
		
		colorChorro = color;
		
	}

	/**
	 * Agrega una nueva gota al chorro (la activa), de todava haber gotas por activar.
	 * De no haber ms gotas que activar, se reinicia la ms antigua a su posicin original.
	 * @param   color   El nuevo color del chorro.
	 **/
	private void agregarGota () {
		
		gotas[indiceGotas++].activar();
		if (indiceGotas == MAX_GOTAS)
			indiceGotas = 0;
		
	}

	/**
	 * Actualiza este chorro.
	 **/
	public void actualizar () {
		
		agregarGota();
		for (Gota g : gotas)
			g.actualizar();
		
	}

	/**
	 * Pinta este chorro.
	 * @param   g   Objeto grfico sobre el cual pintar.
	 **/
	public void pintar (Graphics g) {
		
		for (Gota go : gotas)
			go.pintar(g);
		
	}

// Clases privadas.

	/**
	 * La siguiente clase representa a una gota (una partcula).
	 * Si una gota pasa por debajo de su punto de partida, automticamente se desactiva.
	 **/
	private class Gota {
	
	// Campos de la clase.
	
		/**
		* Radio de la gota.
		**/
		private double radio;
	
		/**
		* Posicin de la gota.
		**/
		private Vector2D pos;
	
		/**
		* Direccin de la gota (ligeramente distinta a la del chorro).
		**/
		private Vector2D dir;
		
		/**
		 * Cantidad de pasos de actualizacin que ha dado la gota.
		 **/
		private int pasos;
	
		/**
		 * Dice si esta gota est activa.
		 **/
		private boolean activa;
	
	// Constructor de la clase.
	
		/**
		 * Constructor de la clase.
		 * Todas las gotas comienzan estando inactivas.
		 **/
		public Gota () {
			
			activar();
			activa = false;
			
		}
	
	// Funciones y metodos de la clase.
	
		/**
		 * Activa esta gota. De ya estar activa, devuelve la gota a su posicin
		 * y direccin de partida. Adems, recalcula un nuevo rado para ella.
		 **/
		public void activar () {
			
			activa = true;
			pos = new Vector2D(posChorro);
			dir = new Vector2D(dirChorro).sumar(new Vector2D((Math.random()*2.0 - 1.0)*EPS, (Math.random()*2.0 - 1.0)*EPS));
			dir = dir.mult(0.9 + Math.random()*0.2);
			radio = radioChorro*(0.5 + Math.random()*0.5);
			pasos = 0;
			
		}
	
		/**
		 * Desactiva esta gota de modo que ya no se actualize o dibuje en pantalla..
		 **/
		public void desactivar () {
			
			activa = false;
			
		}
	
		/**
		* Actualiza esta gota.
		**/
		public void actualizar () {
			
			if (activa) {
				dir.asignarY(dir.obtenerY() + G);
				pos = pos.sumar(dir);
				pasos++;
				if (pos.obtenerY() > posChorro.obtenerY())
					desactivar();
			}
			
		}
	
		/**
		* Pinta esta gota.
		* @param   g   Objeto grfico sobre el cual pintar.
		**/
		public void pintar (Graphics g) {
			
			Vector2D posAnt;
			Color color, color2;
			int i, tamCola;
			
			if (activa) {
				posAnt = new Vector2D(pos);
				tamCola = Math.min(pasos,MAX_VIRTUALES);
				color = MundoDeJuego.obtenerColor(colorChorro);
				color2 = color.darker().darker();
				for (i=0 ; i<tamCola ; i++) {
					posAnt = posAnt.restar(dir.mult(0.5 + Math.random()));
					g.setColor(color2);
					g.fillOval ((int)(posAnt.obtenerX() - radio*1.5), (int)(posAnt.obtenerY() - radio*1.5), (int)(2.0*radio), (int)(2.0*radio));
					g.setColor(color);
					g.fillOval ((int)(posAnt.obtenerX() - radio), (int)(posAnt.obtenerY() - radio), (int)(2.0*radio), (int)(2.0*radio));
				}
				g.setColor(color2);
				g.fillOval ((int)(pos.obtenerX() - radio*1.5), (int)(pos.obtenerY() - radio*1.5), (int)(2.0*radio), (int)(2.0*radio));
				g.setColor(color);
				g.fillOval ((int)(pos.obtenerX() - radio), (int)(pos.obtenerY() - radio), (int)(2.0*radio), (int)(2.0*radio));
			}
			
		}
	
	}

}