import java.util.*;
import java.awt.*;
import javax.sound.midi.*;

/**
 * Esta clase representa la partitura a seguir por el jugador
 * @author Ricardo Monascal
 **/
public class Partitura {

// Constantes de la clase.
	
	/**
	 * Posicion vertical inicial de las partituras.
	 **/
	private static int POS_V = 460;
	
	/**
	 * Tamao total horizontal disponible para la partitura.
	 **/
	private static int TAM_H = 481;
	
	/**
	 * Tamao total vertical disponible para la partitura.
	 **/
	private static int TAM_V = 120;
	
	/**
	 * Desplazamiento para los tamaos de los canales.
	 **/
	private static int DESPL = 5;
	
	/**
	 * Paso de tiempo que se da entre cada actualizacin.
	 **/
	private static int PASO = 4;
	
	/**
	 * Color de fondo para la partitura.
	 **/
	private static Color FONDO = Color.LIGHT_GRAY;
	
	/**
	 * Color para el marcador de color indicado.
	 **/
	private static Color MARCADOR = Color.RED;
	
	/**
	 * Color para el canal que est seleccionado.
	 **/
	private static Color SELECCIONADO = Color.YELLOW;

// Campos de la clase.
	
	/**
	 * Canales de esta partitura
	 **/
	private Canal [] canales;
	
	/**
	 * Tiempo ejecutado en esta partitura.
	 **/
	private int tiempo;
	
	/**
	 * ndice del canal selecionado.
	 **/
	private int seleccionado;
	
	/**
	 * Copia (alias) del reproductor del mundo de juego.
	 **/
	private Reproductor reproductor;

// Constructor de la clase.

	/**
	 * Constructor de la clase.
	 * @param   reproductor   Copia (alias) del reproductor del mundo de juego.
	 **/
	protected Partitura (Reproductor reproductor) {
		
		this.reproductor = reproductor;
		tiempo = -100 * PASO;
		seleccionado = 0;
		
	}

// Funciones y mtodos de la clase.

	/**
	 * Asigna un nuevo numero de canales a esta partitura.
	 * (Corresponde al nivel de dificultad del juego).
	 * @param   numCanales   La nueva cantidad de canales de la partitura.
	 **/
	public void asignarNumCanales (int numCanales) {
		
		int i;
		
		canales = new Canal[numCanales];
		for (i=0 ; i<numCanales ; i++)
			canales[i] = new Canal(i);
		
	}
	
	/**
	 * Devuelve el ndice del canal seleccionado.
	 **/
	public int obtenerSeleccionado () {
		
		return seleccionado;
		
	}
	
	/**
	 * Modifica el ndice del canal seleccionado.
	 * @param   mod   Dice si debe aumentar (true) o disminuir (false).
	 **/
	public void modificarSeleccionado (boolean mod) {
		
		if (mod)
			seleccionado += seleccionado==canales.length-1 ? 0 : 1;
		else
			seleccionado -= seleccionado==0 ? 0 : 1;
		
	}
	
	/**
	 * Calcula las diferencias correspondientes de los colores de las fuentes
	 * con respecto a las partituras.
	 * @param   fuentes   La fuentes del juego
	 **/
	public int sumarDiferencias (Fuente [] fuentes) {
		
		int i, suma, col;
		
		suma = 0;
		for (i=0 ; i<canales.length ; i++) {
			col = canales[i].obtenerColorActivo();
			if (col!=-1 && fuentes[i].obtenerColor()!=canales[i].obtenerColorActivo())
				suma++;
		}
		
		return suma;
		
	}
	
	/**
	 * Deriva una partitura a partir de una secuencia MIDI.
	 * @param   secuencia   Secuencia que se reproducir.
	 **/
	public void derivarPartitura (Sequence secuencia) {
		
		int i, j, paso, ant, ant_col, col, primo, finp;
		
		// Pequeo fragmento de cdigo cableado, jeje.
		paso = 154;
		finp = 8436;
		
		primo = 1;
		for (i=0 ; i<canales.length ; i++) {
			switch (i) {
				case 0:
					primo = 2;
					break;
				case 1:
					primo = 3;
					break;
				case 2:
					primo = 5;
					break;
			}
			canales[i].clear();
			ant = 0;
			ant_col = 0;
			for (j=paso*primo ; j<finp; j+=paso*primo) {
				do {
					col = (int)(Math.random()*4);
				} while (col == ant_col);
				canales[i].add(new Nota(ant, j, col));
				ant = j;
				ant_col = col;
			}
			do {
				col = (int)(Math.random()*4);
			} while (col == ant_col);
			canales[i].add(new Nota(ant, finp, col));
		}
	}
	
	/**
	 * Limpia la partirua (borra la informacin de notas en los canales).
	 **/
	public void limpiar () {
		
		for (Canal c : canales)
			c.clear();
		
		tiempo = -100 * PASO;
		
	}

	/**
	 * Actualiza esta partitura.
	 **/
	public void actualizar () {
		
		if (tiempo == 0)
			reproductor.reproducir();
		for (Canal c : canales)
			c.actualizar();
		tiempo += PASO;
		
	}

	/**
	 * Pinta esta partitura.
	 * @param   g   Objeto grfico sobre el cual pintar.
	 **/
	public void pintar (Graphics g) {
		
		for (Canal c : canales)
			c.pintar(g);
		
	}

// Clases privadas.

	/**
	 * La siguiente clase representa a un canal de la partitura.
	 **/
	private class Canal extends LinkedList<Nota> {
	
	// Campos de la clase
	
		/**
		 * Indica el ndice de ste canal dentro del juego.
		 **/
		private int indice;
	
	// Constructor de la clase.
	
		/**
		 * Constructor de la clase.
		 * @param   indice   ndice de este canal en el juego.
		 **/
		public Canal (int indice) {
			
			super();
			this.indice = indice;
			
		}
	
	// Funciones y mtodos de la clase.

		/**
		* Actualiza este canal.
		**/
		public void actualizar () {
			
			int xFin;
			
			while (!isEmpty()) {
				xFin = TAM_H/2 - getFirst().obtenerFin() + tiempo;
				if (xFin > DESPL + TAM_H)
					removeFirst();
				else
					break;
			}
			
		}	
	
		/**
		* Pinta este canal.
		* @param   g   Objeto grfico sobre el cual pintar.
		**/
		public void pintar (Graphics g) {
			
			int i, tamV, posV, xIni, xFin;
			
			tamV = TAM_V / canales.length;
			posV = POS_V + indice*tamV;
			g.setColor(FONDO);
			g.drawRect(DESPL-3, posV+DESPL-3, TAM_H+5, tamV-DESPL+5);
			if (indice == seleccionado)
				g.setColor(SELECCIONADO);
			g.fillRect(DESPL, posV+DESPL, TAM_H, tamV-DESPL);
			
			g.setFont(g.getFont().deriveFont(24.0f));
			for (Nota n : this) {
				xIni = TAM_H/2 - n.obtenerInicio() + tiempo;
				if (xIni < DESPL)
					break;
				g.setColor(n.obtenerColor());
				xIni = Math.min(DESPL+TAM_H-2, xIni);
				xFin = Math.max(DESPL+2, TAM_H/2 - n.obtenerFin() + tiempo);
				g.fillRect(xFin, posV+DESPL+2, xIni-xFin, tamV-DESPL-4);
				if (xIni-xFin > 3*DESPL) {
					g.setColor(MundoDeJuego.obtenerColor((n.obtenerIndice()+1)%4));
					g.drawString(""+(n.obtenerIndice()+1), xIni-3*DESPL, posV+tamV/2+12);
				}
			}
			
			if (indice == seleccionado)
				g.setColor(Color.GREEN);
			else
				g.setColor(MARCADOR);
			g.drawLine(TAM_H/2 - 1, posV, TAM_H/2 - 1, posV+tamV+DESPL-2);
			g.drawLine(TAM_H/2 - 2, posV-1, TAM_H/2 - 2, posV+tamV+DESPL-1);
			g.drawLine(TAM_H/2 + 2, posV-1, TAM_H/2 + 2, posV+tamV+DESPL-1);
			g.drawLine(TAM_H/2 + 1, posV, TAM_H/2 + 1, posV+tamV+DESPL-2);
			
			if (indice == seleccionado) {
				g.drawLine(TAM_H/2 - 5, posV, TAM_H/2 - 5, posV+tamV+DESPL-2);
				g.drawLine(TAM_H/2 + 5, posV, TAM_H/2 + 5, posV+tamV+DESPL-2);
			}
			
		}
	
		/**
		 * Obtiene el indice del color que est activo de este canal.
		 * De no haberlo, devuelve -1.
		 **/
		public int obtenerColorActivo () {
			
			int xIni, xFin;
			
			for (Nota n : this) {
				xIni = TAM_H/2 - n.obtenerInicio() + tiempo;
				if (xIni < DESPL)
					break;
				xFin = TAM_H/2 - n.obtenerFin() + tiempo;
				if (xIni >= TAM_H/2 && xFin < TAM_H/2)
					return n.obtenerIndice();
			}
			
			return -1;
			
		}
		
	}

	/**
	 * La siguiente clase representa a una nota particular con inicio, final y color.
	 **/
	private class Nota {
	
	// Campos de la clase.
	
		/**
		 * Inicio de la nota (en unidades de tiempo, dependientes del fps).
		 **/
		private int inicio;
	
		/**
		 * Fin de la nota (en unidades de tiempo, dependientes del fps).
		 **/
		private int fin;
	
		/**
		 * Color de la nota (indice en el arreglo de colores del mundo de juego).
		 **/
		private int color;
	
	// Constructor de la clase.
	
		/**
		 * Constructor de la clase.
		 * @param   inicio   Inicio de la nota.
		 * @param   fin   Inicio de la nota.
		 * @param   color   Inicio de la nota.
		 **/
		public Nota (int inicio, int fin, int color) {
			
			this.inicio = inicio;
			this.fin = fin;
			this.color = color;
			
		}
	
	// Funciones y mtodos de la clase.
	
		/**
		 * Obtiene el inicio de la nota.
		 **/
		public int obtenerInicio () {
			
			return inicio;
			
		}
	
		/**
		 * Obtiene el fin de la nota.
		 **/
		public int obtenerFin () {
			
			return fin;
			
		}
	
		/**
		 * Obtiene color de la nota. (Buscando el color indicado en el
		 * arreglo de colores del mundo de juego).
		 **/
		public Color obtenerColor () {
			
			return MundoDeJuego.obtenerColor(color);
			
		}
	
		/**
		 * Obtiene el indice del color de la nota..
		 **/
		public int obtenerIndice () {
			
			return color;
			
		}
	
	}

}